@@ -3,49 +3,60 @@
 # region #-- imports --#
 from __future__ import annotations
 
+import aiohttp
+import asyncio
 import base64
 import json
 import logging
-from asyncio import TimeoutError
 from typing import (
+    Any,
     Dict,
     List,
     Optional,
 )
 
-import aiohttp
-from aiohttp.client_exceptions import (
-    ClientConnectionError,
-    ClientConnectorError,
-    ContentTypeError,
-)
-
 from .exceptions import (
+    MeshBadResponse,
     MeshConnectionError,
+    MeshInvalidCredentials,
+    MeshInvalidInput,
+    MeshInvalidOutput,
+    MeshNodeNotPrimary,
     MeshTimeoutError,
 )
 from .logger import LoggerFormatter
 
 # endregion
 
-
 _LOGGER = logging.getLogger(__name__)
 
 
+def jnap_url(target) -> str:
+    """Return the URL that should be used for the request
+
+    :param target: the API host
+    :return: string containing the base URL for all JNAP requests
+    """
+
+    # noinspection HttpUrlsUsage
+    return f"http://{target}/JNAP/"
+
+
 class Actions:
     """Represents the available actions"""
 
     # noinspection HttpUrlsUsage
     ROOT: str = "http://linksys.com/jnap"
-    
+
     CHECK_PASSWORD: str = f"{ROOT}/core/CheckAdminPassword"
     DELETE_DEVICE: str = f"{ROOT}/devicelist/DeleteDevice"
     GET_BACKHAUL: str = f"{ROOT}/nodes/diagnostics/GetBackhaulInfo"
     GET_DEVICES: str = f"{ROOT}/devicelist/GetDevices3"
+    GET_FIRMWARE_UPDATE_SETTINGS: str = f"{ROOT}/firmwareupdate/GetFirmwareUpdateSettings"
     GET_GUEST_NETWORK_INFO: str = f"{ROOT}/guestnetwork/GetGuestRadioSettings2"
     GET_PARENTAL_CONTROL_INFO: str = f"{ROOT}/parentalcontrol/GetParentalControlSettings"
     GET_SPEEDTEST_RESULTS: str = f"{ROOT}/healthcheck/GetHealthCheckResults"
-    GET_SPEEDTEST_STATE: str = f"{ROOT}/healthcheck/GetHealthCheckStatus"
+    GET_SPEEDTEST_STATUS: str = f"{ROOT}/healthcheck/GetHealthCheckStatus"
     GET_STORAGE_PARTITIONS: str = f"{ROOT}/nodes/storage/GetNodesPartitions"
     GET_STORAGE_SMB_SERVER: str = f"{ROOT}/nodes/storage/GetSMBServerSettings"
     GET_UPDATE_FIRMWARE_STATE: str = f"{ROOT}/nodes/firmwareupdate/GetFirmwareUpdateStatus"
@@ -59,8 +70,20 @@ class Actions:
     UPDATE_FIRMWARE: str = f"{ROOT}/nodes/firmwareupdate/UpdateFirmwareNow"
 
 
+class Defaults:
+    """Represents the default payloads required for requests"""
+
+    PAYLOADS: Dict[str, Dict] = {
+        Actions.GET_SPEEDTEST_RESULTS: {
+            "healthCheckModule": "SpeedTest",
+            "includeModuleResults": True,
+            "lastNumberOfResults": 1,
+        },
+    }
+
+
 class Request(LoggerFormatter):
-    """Represents a request to be made to the API"""
+    """Represents a request for the API"""
 
     def __init__(
         self,
@@ -83,24 +106,13 @@ class Request(LoggerFormatter):
 
         super().__init__(prefix=f"{self.__class__.__name__}.")
 
-        self._action: str | List[str] = action
+        self._action: str = action
         self._creds: str = base64.b64encode(bytes(f"{username}:{password}", "utf-8")).decode("ascii")
         self._payload: Optional[List[Dict], Dict] = payload
         self._session: Optional[aiohttp.ClientSession] = session or aiohttp.ClientSession(raise_for_status=True)
         self._target: str = target
 
-        self._jnap_url: str = self.jnap_url(target=self._target)
-
-    @staticmethod
-    def jnap_url(target) -> str:
-        """Return the URL that should be used for the request
-
-        :param target: the API host
-        :return: string containing the base URL for all JNAP requests
-        """
-
-        # noinspection HttpUrlsUsage
-        return f"http://{target}/JNAP/"
+        self._jnap_url: str = jnap_url(target=self._target)
 
     async def execute(self, timeout: int = 10) -> Response:
         """Send the request
@@ -133,9 +145,9 @@ class Request(LoggerFormatter):
                 timeout=timeout
             )
             resp_json = await resp.json()
-        except TimeoutError:
+        except asyncio.TimeoutError:
             raise MeshTimeoutError
-        except (ClientConnectionError, ClientConnectorError, ContentTypeError,):
+        except (aiohttp.ClientConnectionError, aiohttp.ClientConnectorError, aiohttp.ContentTypeError,):
             raise MeshConnectionError from None
         except json.JSONDecodeError as err:
             raise err from None
@@ -147,13 +159,11 @@ class Request(LoggerFormatter):
 class Response(LoggerFormatter):
     """Represents a response from the API"""
 
-    ACTION_KEY: str = "action"
+    DATA_KEY_SINGLE: str = "output"
+    DATA_KEY_TRANSACTION: str = "responses"
     RESULT_KEY: str = "result"
-    RESULTS_KEY_SINGLE: str = "output"
-    RESULTS_KEY_TRANSACTION: str = "responses"
-    RESULTS_KEY_ERROR: str = "error"
 
-    def __init__(self, action: str, data: Dict) -> None:
+    def __init__(self, action: str, data: Dict[str, Any]) -> None:
         """Constructor
 
         :param action: The action that was issued in the request to cause the response
@@ -163,45 +173,69 @@ class Response(LoggerFormatter):
         super().__init__(prefix=f"{self.__class__.__name__}.")
 
         self._action: str = action
-        self._data: Dict = data
+        self._data: Dict[str, Any] = data
 
+        self._process_data()
+
+    def _process_data(self) -> None:
+        """Process the given data to check for errors"""
+
+        if self._data.get(self.RESULT_KEY) != "OK":
+            responses = (
+                self.data
+                if self.action == Actions.TRANSACTION
+                else [self.data]
+            )
+
+            err = None
+            for resp in responses:
+                err = None
+                if resp.get(self.RESULT_KEY) == "_ErrorInvalidInput":
+                    err = MeshInvalidInput(resp.get("error"))
+                elif resp.get(self.RESULT_KEY) == "_ErrorInvalidOutput":
+                    err = MeshInvalidOutput(resp.get("error"))
+                elif resp.get(self.RESULT_KEY) == "_ErrorUnauthorized":
+                    err = MeshInvalidCredentials
+                elif resp.get(self.RESULT_KEY) == "_ErrorUnknownAction":
+                    action = (
+                        resp.get("error")
+                        if self.action == Actions.TRANSACTION
+                        else f"Unknown action URI '{self.action}'"
+                    )
+                    err = MeshInvalidInput(action)
+                elif resp.get(self.RESULT_KEY) == "ErrorDeviceNotInMasterMode":
+                    err = MeshNodeNotPrimary
+                elif resp.get(self.RESULT_KEY).startswith("_"):
+                    err = MeshInvalidInput(resp.get(self.RESULT_KEY))
+
+                if err:
+                    break
+
+            if err is None:
+                _LOGGER.error(self.message_format("unknown error received: %s"), json.dumps(self._data))
+                err = MeshBadResponse
+
+            raise err
+
+    # region #-- properties --#
     @property
     def action(self) -> str:
-        """Return the action associated with the response
+        """Return the action that resulted in the response
 
-        :return: the JNAP action used in the request
+        :return: string containing the action
         """
 
         return self._action
 
     @property
-    def data(self) -> Dict:
-        """Return the data as returned by the API
+    def data(self) -> Dict[str, Any]:
+        """"""
 
-        :return: If a successful request the return will be a dictionary
-                 in the form {"action": <action_for_request>, "result": <result_as_per_API>}
-                 If the request was unsuccessfull the return will be in the form
-                 {"result": <result_as_per_API>}
-        """
-
-        ret: Dict = {
-            self.ACTION_KEY: self._action
-        }
-        if self.is_successful:
-            ret[self.RESULT_KEY] = (
-                self._data.get(self.RESULTS_KEY_SINGLE)
-                or self._data.get(self.RESULTS_KEY_TRANSACTION)
-            )
-        else:
-            ret[self.RESULT_KEY] = self._data
+        ret = (
+            self._data.get(self.DATA_KEY_TRANSACTION)
+            if self.action == Actions.TRANSACTION
+            else self._data.get(self.DATA_KEY_SINGLE)
+        )
 
         return ret
-
-    @property
-    def is_successful(self) -> bool:
-        """Check if the response indicates a successful request
-
-        :return: True if successful otherwise False
-        """
-
-        return self._data.get(self.RESULT_KEY) == "OK"
+    # endregion
